//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Pol/PolFilter.cpp
//! @brief     Implements class DetectionProperties.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Pol/PolFilter.h"
#include "Fit/Param/RealLimits.h"
#include <heinz/Complex.h>

PolFilter::PolFilter(R3 direction, double efficiency, double total_transmission)
    : m_direction(direction), m_efficiency(efficiency), m_total_transmission(total_transmission)
{
    RealLimits::nonnegative().check("Transmission", m_total_transmission);
}

PolFilter::PolFilter() : PolFilter({}, {}, 1.0) {}

PolFilter::PolFilter(const PolFilter& other)
    : PolFilter(other.m_direction, other.m_efficiency, other.m_total_transmission)
{
}

void PolFilter::setDirEffTra(const R3 direction, double efficiency, double total_transmission)
{
    if (!check(direction, efficiency, total_transmission))
        throw std::runtime_error("IDetector2D::setAnalyzer: the "
                                 "given properties are not physical");
    if (efficiency == 0.0 || total_transmission == 0.0 || direction.mag() == 0.0) {
        m_direction = R3{};
        m_efficiency = 0.0;
    } else {
        m_direction = direction.unit();
        m_efficiency = efficiency;
    }
    m_total_transmission = total_transmission;
}

Eigen::Matrix2cd PolFilter::matrix() const
{
    if (m_direction.mag() == 0.0 || m_efficiency == 0.0)
        return m_total_transmission * Eigen::Matrix2cd::Identity();
    Eigen::Matrix2cd result;
    double x = m_direction.x() / m_direction.mag();
    double y = m_direction.y() / m_direction.mag();
    double z = m_direction.z() / m_direction.mag();
    double sum = m_total_transmission * 2.0;
    double diff = m_total_transmission * m_efficiency * 2.0;
    complex_t im(0.0, 1.0);
    result(0, 0) = (sum + diff * z) / 2.0;
    result(0, 1) = diff * (x - im * y) / 2.0;
    result(1, 0) = diff * (x + im * y) / 2.0;
    result(1, 1) = (sum - diff * z) / 2.0;
    return result;
}

R3 PolFilter::polDirection() const
{
    return m_direction;
}

double PolFilter::polEfficiency() const
{
    return m_efficiency;
}

double PolFilter::totalTransmission() const
{
    return m_total_transmission;
}

bool PolFilter::check(const R3 direction, double efficiency, double total_transmission) const
{
    if (direction.mag() == 0.0)
        return false;
    double aplus = total_transmission * (1.0 + efficiency);
    double amin = total_transmission * (1.0 - efficiency);
    if (aplus < 0.0 || aplus > 1.0)
        return false;
    if (amin < 0.0 || amin > 1.0)
        return false;
    return true;
}
