//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Fit/ParameterTreeItems.h
//! @brief     Defines classes for ParameterTreeItems
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_FIT_PARAMETERTREEITEMS_H
#define BORNAGAIN_GUI_MODEL_FIT_PARAMETERTREEITEMS_H

#include "GUI/Model/Session/SessionItem.h"
#include "GUI/Model/Types/DoubleDescriptor.h"

class MaterialItem;

//! ParameterTreeItems is a collection of items necessary to form a tuning tree for
//! real time widget.

//! The ParameterLabelItem class represents a label (string without value, like 'Layer',
//! 'MultiLayer') in a parameter tuning tree.

class ParameterLabelItem : public SessionItem {
public:
    static constexpr auto M_TYPE{"Parameter Label"};

    ParameterLabelItem();
};

//! The ParameterItem class represent a tuning value in a parameter tuning tree.

class ParameterItem : public SessionItem {
public:
    static constexpr auto M_TYPE{"Parameter"};

    ParameterItem();

    void propagateValueToLink(double newValue);

    //! Unique string to identify this ParameterItem.
    //!
    //! The link is arbitrary. It can't be used for finding the linked item (therefore it does
    //! not have to be a model path). However, it is used for comparison, also across project
    //! load/save. Therefore the link is always the same, not e.g. an always generated UUID.
    //! This link is used for setting backup values and for finding this ParameterItem when
    //! referring from fit parameters.
    QString link() const;

    //! Links this item to the given value defined by a descriptor.
    void linkToDescriptor(DoubleDescriptor d);

    //! Links this item to the given session item.
    void linkToSessionItem(SessionItem* item);

    RealLimits limitsOfLink() const;
    int decimalsOfLink() const;

private:
    QString m_link;       //!< See docu of link()
    DoubleDescriptor m_d; //!< The linked double value
};

//! The ParameterContainerItem is a top item to hold all ParameterItem, represents an entry
//! point to parameter tuning tree. Part of JobItem.

class ParameterContainerItem : public SessionItem {
public:
    static constexpr auto M_TYPE{"Parameter Container"};

    ParameterContainerItem();

    void writeNonSessionItems(QXmlStreamWriter* writer) const override;
    void readNonSessionItems(QXmlStreamReader* reader) override;
    bool allowWritingChildToXml(SessionItem* child) const override;

    void setBackupValue(const QString& link, double d);
    void restoreBackupValues();

private:
    void restoreBackupValue(SessionItem* item);

    QMap<QString, double> m_backupValues;
};

#endif // BORNAGAIN_GUI_MODEL_FIT_PARAMETERTREEITEMS_H
