//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Instrument/FootprintItems.cpp
//! @brief     Implements FootprintItem classes
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Instrument/FootprintItems.h"
#include "Device/Beam/FootprintGauss.h"
#include "Device/Beam/FootprintSquare.h"

// Base class
/* ------------------------------------------------ */

FootprintItem::FootprintItem(const QString& model_type) : SessionItem(model_type) {}

FootprintItem::~FootprintItem() = default;

// Footprint none
/* ------------------------------------------------ */

FootprintNoneItem::FootprintNoneItem() : FootprintItem(M_TYPE) {}

FootprintNoneItem::~FootprintNoneItem() = default;

std::unique_ptr<IFootprintFactor> FootprintNoneItem::createFootprint() const
{
    return {};
}

// Gaussian footprint
/* ------------------------------------------------ */ // sic!

FootprintGaussianItem::FootprintGaussianItem() : FootprintItem(M_TYPE)
{
    addProperty(P_VALUE, 0.0)
        ->setLimits(RealLimits::nonnegative())
        .setToolTip("The ratio of beam and sample full widths");
}

FootprintGaussianItem::~FootprintGaussianItem() = default;

std::unique_ptr<IFootprintFactor> FootprintGaussianItem::createFootprint() const
{
    return std::make_unique<FootprintGauss>(getItemValue(P_VALUE).toDouble());
}

double FootprintGaussianItem::gaussianFootprintValue() const
{
    return getItemValue(P_VALUE).toDouble();
}

void FootprintGaussianItem::setGaussianFootprintValue(double value)
{
    setItemValue(P_VALUE, value);
}

// Square footprint
/* ------------------------------------------------ */

FootprintSquareItem::FootprintSquareItem() : FootprintItem(M_TYPE)
{
    addProperty(P_VALUE, 0.0)
        ->setLimits(RealLimits::nonnegative())
        .setToolTip("The ratio of beam and sample full widths");
}

FootprintSquareItem::~FootprintSquareItem() = default;

std::unique_ptr<IFootprintFactor> FootprintSquareItem::createFootprint() const
{
    return std::make_unique<FootprintSquare>(getItemValue(P_VALUE).toDouble());
}

double FootprintSquareItem::squareFootprintValue() const
{
    return getItemValue(P_VALUE).toDouble();
}

void FootprintSquareItem::setSquareFootprintValue(double value)
{
    setItemValue(P_VALUE, value);
}
