//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Instrument/SpecularBeamInclinationItem.cpp
//! @brief     Implements class SpecularBeamInclinationItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Instrument/SpecularBeamInclinationItem.h"
#include "Base/Const/Units.h"
#include "GUI/Model/Data/PointwiseAxisItem.h"
#include "GUI/Model/Group/GroupItem.h"

namespace {

void setupDistributionMean(SessionItem* distribution);
void setAxisPresentationDefaults(SessionItem* axis_item, const QString& type);

GroupInfo createAxesGroup()
{
    GroupInfo info;
    info.add(BasicAxisItem::M_TYPE, "Uniform axis");
    info.add(PointwiseAxisItem::M_TYPE, "Non-uniform axis");
    info.setDefaultType(BasicAxisItem::M_TYPE);
    return info;
}

} // namespace

SpecularBeamInclinationItem::SpecularBeamInclinationItem()
    : BeamDistributionItem(M_TYPE, m_show_mean)
{
    register_distribution_group(Distributions::Symmetric);
    setupAxisGroup();
    setupDistributionMean(distribution());

    initDistributionItem(m_show_mean);
}

SpecularBeamInclinationItem::~SpecularBeamInclinationItem() = default;

double SpecularBeamInclinationItem::scaleFactor() const
{
    return Units::deg;
}

void SpecularBeamInclinationItem::updateFileName(const QString& filename)
{
    auto* group_item = item<GroupItem>(P_ALPHA_AXIS);
    auto* axis_item = group_item->firstChildOfType<PointwiseAxisItem>();
    ASSERT(axis_item);
    axis_item->setFileName(filename);
}

BasicAxisItem* SpecularBeamInclinationItem::alphaAxis() const
{
    return dynamic_cast<BasicAxisItem*>(getGroupItem(P_ALPHA_AXIS));
}

GroupItem* SpecularBeamInclinationItem::alphaAxisItem() const
{
    return item<GroupItem>(P_ALPHA_AXIS);
}

void SpecularBeamInclinationItem::setupAxisGroup()
{
    static const GroupInfo axesGroup = createAxesGroup();
    auto* group_item = dynamic_cast<GroupItem*>(this->addGroupProperty(P_ALPHA_AXIS, axesGroup));

    // Both underlying axis items are created, since it
    // strongly simplifies their representation and state
    // handling (no signal emulation required).
    // Basic axis item is the default one.

    group_item->setCurrentType(PointwiseAxisItem::M_TYPE);
    setAxisPresentationDefaults(group_item->currentItem(), group_item->currentType());

    group_item->setCurrentType(BasicAxisItem::M_TYPE);
    setAxisPresentationDefaults(group_item->currentItem(), group_item->currentType());

    group_item->setToolTip("Axis type selected");
    group_item->setDisplayName("Axis type");
    group_item->setEnabled(false);
    group_item->mapper()->setOnValueChange(
        [group_item]() {
            if (group_item->currentItem()->modelType() == PointwiseAxisItem::M_TYPE)
                group_item->setEnabled(true);
        },
        this);
}

namespace {

void setupDistributionMean(SessionItem* distribution)
{
    ASSERT(distribution);

    SessionItem* valueItem = dynamic_cast<SymmetricDistributionItem*>(distribution)->meanItem();
    ASSERT(valueItem);

    valueItem->setLimits(RealLimits::limited(-90.0, 90.0));
    valueItem->setDecimals(3);
    valueItem->setValue(0.0);
}

void setAxisPresentationDefaults(SessionItem* item, const QString& type)
{
    auto* axis_item = dynamic_cast<BasicAxisItem*>(item);
    Q_ASSERT(axis_item);

    axis_item->titleItem()->setVisible(false);
    axis_item->setTitle("alpha_i");
    axis_item->binsItem()->setToolTip("Number of points in scan");
    axis_item->lowerBoundItem()->setToolTip("Starting value [deg]");
    axis_item->upperBoundItem()->setToolTip("Ending value [deg]");
    axis_item->lowerBoundItem()->setLimits(RealLimits::limited(0., 90.));
    axis_item->upperBoundItem()->setLimits(RealLimits::limited(0., 90.));

    if (type == BasicAxisItem::M_TYPE) {
        axis_item->setLowerBound(0.0);
        axis_item->setUpperBound(3.0);
        axis_item->setBinCount(500);
    } else if (type == PointwiseAxisItem::M_TYPE) {
        axis_item->lowerBoundItem()->setEnabled(false);
        axis_item->upperBoundItem()->setEnabled(false);
        axis_item->binsItem()->setEnabled(false);
    }
}

} // namespace
