//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Trafo/RotationItems.cpp
//! @brief     Implements class RotationItems
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Trafo/RotationItems.h"
#include "Base/Const/Units.h"
#include "Base/Vector/RotMatrix.h"
#include "GUI/Model/Types/DoubleDescriptor.h"
#include "Sample/Scattering/Rotations.h"

using namespace Units;

// ----------------------------------------------------------------------------

RotationItem::RotationItem(const QString& name) : SessionItem(name) {}

RotMatrix RotationItem::rotation() const
{
    auto p = createRotation();
    return (p != nullptr) ? p->getRotMatrix() : RotMatrix();
}

// ----------------------------------------------------------------------------

XRotationItem::XRotationItem() : RotationItem(M_TYPE)
{
    setToolTip("Particle rotation around x-axis");
    addProperty(P_ANGLE, 0.0)->setToolTip("Rotation angle around x-axis in degrees");
}

unique_ptr<IRotation> XRotationItem::createRotation() const
{
    return std::make_unique<RotationX>(deg2rad(angle()));
}

DoubleDescriptor XRotationItem::angle() const
{
    DoubleDescriptor d(getItem(P_ANGLE), Unit::degree);
    d.tooltip = "Rotation angle around x-axis"; // tooltip w/o unit
    return d;
}

// ----------------------------------------------------------------------------

YRotationItem::YRotationItem() : RotationItem(M_TYPE)
{
    setToolTip("Particle rotation around y-axis");
    addProperty(P_ANGLE, 0.0)->setToolTip("Rotation angle around y-axis in degrees");
}

unique_ptr<IRotation> YRotationItem::createRotation() const
{
    return std::make_unique<RotationY>(deg2rad(angle()));
}

DoubleDescriptor YRotationItem::angle() const
{
    DoubleDescriptor d(getItem(P_ANGLE), Unit::degree);
    d.tooltip = "Rotation angle around y-axis"; // tooltip w/o unit
    return d;
}

// ----------------------------------------------------------------------------

ZRotationItem::ZRotationItem() : RotationItem(M_TYPE)
{
    setToolTip("Particle rotation around z-axis");
    addProperty(P_ANGLE, 0.0)->setToolTip("Rotation angle around z-axis in degrees");
}

unique_ptr<IRotation> ZRotationItem::createRotation() const
{
    return std::make_unique<RotationZ>(deg2rad(angle()));
}

DoubleDescriptor ZRotationItem::angle() const
{
    DoubleDescriptor d(getItem(P_ANGLE), Unit::degree);
    d.tooltip = "Rotation angle around z-axis"; // tooltip w/o unit
    return d;
}

// ----------------------------------------------------------------------------

EulerRotationItem::EulerRotationItem() : RotationItem(M_TYPE)
{
    setToolTip("Sequence of three rotations following Euler angles \n"
               "notation z-x'-z'");
    addProperty(P_ALPHA, 0.0)->setToolTip("First Euler angle in z-x'-z' sequence in degrees");
    addProperty(P_BETA, 0.0)->setToolTip("Second Euler angle in z-x'-z' sequence in degrees");
    addProperty(P_GAMMA, 0.0)->setToolTip("Third Euler angle in z-x'-z' sequence in degrees");
}

unique_ptr<IRotation> EulerRotationItem::createRotation() const
{
    return std::make_unique<RotationEuler>(deg2rad(alpha()), deg2rad(beta()), deg2rad(gamma()));
}

DoubleDescriptor EulerRotationItem::alpha() const
{
    DoubleDescriptor d(getItem(P_ALPHA), Unit::degree);
    d.tooltip = "First Euler angle in z-x'-z' sequence"; // tooltip w/o unit
    return d;
}

DoubleDescriptor EulerRotationItem::beta() const
{
    DoubleDescriptor d(getItem(P_BETA), Unit::degree);
    d.tooltip = "Second Euler angle in z-x'-z' sequence"; // tooltip w/o unit
    return d;
}

DoubleDescriptor EulerRotationItem::gamma() const
{
    DoubleDescriptor d(getItem(P_GAMMA), Unit::degree);
    d.tooltip = "Third Euler angle in z-x'-z' sequence"; // tooltip w/o unit
    return d;
}
