//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/FootprintCorrectionEditor.cpp
//! @brief     Implements class FootprintCorrectionEditor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/FootprintCorrectionEditor.h"
#include "GUI/Model/Instrument/FootprintItems.h"
#include "GUI/Model/Instrument/InstrumentItems.h"

#include <QComboBox>
#include <QDoubleSpinBox>
#include <QFormLayout>
#include <QGroupBox>
#include <QStackedLayout>

//==================================================================================================
// SquareFootprintForm
//==================================================================================================

class SquareFootprintForm : public QWidget {
public:
    SquareFootprintForm(QWidget* parent = nullptr, Qt::WindowFlags f = Qt::WindowFlags());
    void setFootprint(FootprintSquareItem* item);

private slots:
    void onValueChanged(double value);

private:
    QDoubleSpinBox* m_spinbox;
    FootprintSquareItem* m_item;
};

//--------------------------------------------------------------------------------------------------
// public member functions
//--------------------------------------------------------------------------------------------------

SquareFootprintForm::SquareFootprintForm(QWidget* parent, Qt::WindowFlags f)
    : QWidget(parent, f), m_item(nullptr)
{
    auto* layout = new QFormLayout(this);
    layout->setMargin(0);

    m_spinbox = new QDoubleSpinBox(this);
    connect(m_spinbox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &SquareFootprintForm::onValueChanged);
    m_spinbox->setDecimals(3);
    m_spinbox->setRange(0, std::numeric_limits<double>::max());
    m_spinbox->setSingleStep(0.01);
    m_spinbox->setToolTip("The ratio of beam and sample full widths");
    layout->addRow("Width ratio:", m_spinbox);
}

void SquareFootprintForm::setFootprint(FootprintSquareItem* item)
{
    m_item = item;
    setEnabled(m_item);
    if (m_item)
        m_spinbox->setValue(m_item->squareFootprintValue());
    else
        m_spinbox->setValue(0);
}

//--------------------------------------------------------------------------------------------------
// private slots
//--------------------------------------------------------------------------------------------------

void SquareFootprintForm::onValueChanged(double value)
{
    if (m_item)
        m_item->setSquareFootprintValue(value);
}

//==================================================================================================
// GaussFootprintForm
//==================================================================================================

class GaussFootprintForm : public QWidget {
public:
    GaussFootprintForm(QWidget* parent = nullptr, Qt::WindowFlags f = Qt::WindowFlags());
    void setFootprint(FootprintGaussianItem* item);

private slots:
    void onValueChanged(double value);

private:
    QDoubleSpinBox* m_spinbox;
    FootprintGaussianItem* m_item;
};

//--------------------------------------------------------------------------------------------------
// public member functions
//--------------------------------------------------------------------------------------------------

GaussFootprintForm::GaussFootprintForm(QWidget* parent, Qt::WindowFlags f)
    : QWidget(parent, f), m_item(nullptr)
{
    auto* layout = new QFormLayout(this);
    layout->setMargin(0);

    m_spinbox = new QDoubleSpinBox(this);
    connect(m_spinbox, qOverload<double>(&QDoubleSpinBox::valueChanged), this,
            &GaussFootprintForm::onValueChanged);
    m_spinbox->setDecimals(3);
    m_spinbox->setRange(0, std::numeric_limits<double>::max());
    m_spinbox->setSingleStep(0.01);
    m_spinbox->setToolTip("The ratio of beam and sample full widths");
    layout->addRow("Width ratio:", m_spinbox);
}

void GaussFootprintForm::setFootprint(FootprintGaussianItem* item)
{
    m_item = item;
    setEnabled(m_item);
    if (m_item)
        m_spinbox->setValue(m_item->gaussianFootprintValue());
    else
        m_spinbox->setValue(0);
}

//--------------------------------------------------------------------------------------------------
// private slots
//--------------------------------------------------------------------------------------------------

void GaussFootprintForm::onValueChanged(double value)
{
    if (m_item)
        m_item->setGaussianFootprintValue(value);
}

//==================================================================================================
// FootprintCorrectionEditor
//==================================================================================================

//--------------------------------------------------------------------------------------------------
// public member functions
//--------------------------------------------------------------------------------------------------

FootprintCorrectionEditor::FootprintCorrectionEditor(QWidget* parent, Qt::WindowFlags f)
    : QWidget(parent, f), m_item(nullptr)
{
    auto* layout = new QVBoxLayout(this);

    auto* gbox = new QGroupBox("Footprint correction", this);
    layout->addWidget(gbox);
    auto* form = new QFormLayout(gbox);

    m_combo = new QComboBox(gbox);
    form->addRow("Type:", m_combo);

    m_stack = new QStackedLayout();
    form->addRow(m_stack);

    m_combo->addItem("None", FootprintNoneItem::M_TYPE);
    m_combo->addItem("Square", FootprintSquareItem::M_TYPE);
    m_combo->addItem("Gaussian", FootprintGaussianItem::M_TYPE);
    connect(m_combo, qOverload<int>(&QComboBox::currentIndexChanged), this,
            &FootprintCorrectionEditor::onFootprintSelected);

    m_blankForm = new QWidget(gbox);
    m_stack->addWidget(m_blankForm);

    m_squareForm = new SquareFootprintForm(gbox);
    m_stack->addWidget(m_squareForm);

    m_gaussForm = new GaussFootprintForm(gbox);
    m_stack->addWidget(m_gaussForm);
}

void FootprintCorrectionEditor::setItem(SpecularBeamItem* item)
{
    m_item = item;
    setEnabled(m_item);
    if (m_item) {
        int idx = m_combo->findData(m_item->footprint()->modelType());
        ASSERT(idx != -1);
        if (idx != m_combo->currentIndex())
            m_combo->setCurrentIndex(idx);
        else if (!m_item->footprint()->hasModelType<FootprintNoneItem>())
            onFootprintSelected(idx); // trigger update manually
    }
}

void FootprintCorrectionEditor::clearItem()
{
    setItem(nullptr);
}

//--------------------------------------------------------------------------------------------------
// private slots
//--------------------------------------------------------------------------------------------------

void FootprintCorrectionEditor::onFootprintSelected(int index)
{
    if (m_item) {
        const QString footprint = m_combo->itemData(index).toString();
        if (footprint == FootprintSquareItem::M_TYPE) {
            m_squareForm->setFootprint(m_item->setFootprintType<FootprintSquareItem>());
            m_stack->setCurrentWidget(m_squareForm);
        } else if (footprint == FootprintGaussianItem::M_TYPE) {
            m_gaussForm->setFootprint(m_item->setFootprintType<FootprintGaussianItem>());
            m_stack->setCurrentWidget(m_gaussForm);
        } else if (footprint == FootprintNoneItem::M_TYPE) {
            m_item->setFootprintType<FootprintNoneItem>();
            m_stack->setCurrentWidget(m_blankForm);
        } else
            ASSERT(false);
    }
}
