//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/SpecularBeamEditor.cpp
//! @brief     Defines class SpecularBeamEditor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/SpecularBeamEditor.h"
#include "GUI/Model/Instrument/BeamItems.h"
#include "GUI/Model/Instrument/BeamWavelengthItem.h"
#include "GUI/View/Instrument/DistributionEditor.h"
#include "GUI/View/Instrument/FootprintCorrectionEditor.h"
#include "GUI/View/Instrument/InclinationAnglesEditor.h"

#include <QDoubleValidator>
#include <QFormLayout>
#include <QGridLayout>
#include <QLineEdit>

//==================================================================================================
// SpecularBeamEditor
//==================================================================================================

//--------------------------------------------------------------------------------------------------
// public member functions
//--------------------------------------------------------------------------------------------------

SpecularBeamEditor::SpecularBeamEditor(QWidget* parent, Qt::WindowFlags f)
    : QWidget(parent, f), m_item(nullptr)

{
    auto* grid = new QGridLayout(this);
    grid->setColumnStretch(0, 1);
    grid->setColumnStretch(1, 1);
    grid->setColumnStretch(2, 1);

    auto* form = new QFormLayout();
    grid->addLayout(form, 0, 0);
    m_intensityLineEdit = new QLineEdit(this);
    m_intensityLineEdit->setToolTip("Beam intensity in neutrons (or gammas per second.");
    auto* validator = new QDoubleValidator(m_intensityLineEdit);
    validator->setNotation(QDoubleValidator::ScientificNotation);
    validator->setRange(0.0, 1e+32, 1000);
    m_intensityLineEdit->setValidator(validator);
    connect(m_intensityLineEdit, &QLineEdit::editingFinished, this,
            &SpecularBeamEditor::onIntensityEditingFinished);
    form->addRow("Intensity:", m_intensityLineEdit);

    // valid input range is ]0,inf[
    MeanConfig wave_cfg{boost::numeric::interval<double>(std::numeric_limits<double>::min(),
                                                         std::numeric_limits<double>::max()),
                        4, true};
    m_wavelengthEditor = new DistributionEditor("Wavelength [nm]", wave_cfg,
                                                GUI::ID::Distributions::Symmetric, this);
    grid->addWidget(m_wavelengthEditor, 1, 0);

    m_inclinationEditor = new InclinationAnglesEditor(this);
    grid->addWidget(m_inclinationEditor, 1, 1);

    m_footprintEditor = new FootprintCorrectionEditor(this);
    grid->addWidget(m_footprintEditor, 1, 2);
}

void SpecularBeamEditor::setItem(SpecularBeamItem* item)
{
    m_item = item;
    setEnabled(m_item);
    if (m_item) {
        m_intensityLineEdit->setText(QString::number(m_item->intensity()));
        m_wavelengthEditor->setItem(m_item->wavelengthItem());
        m_inclinationEditor->setItem(m_item->inclinationAngleItem());
        m_footprintEditor->setItem(m_item);
    } else {
        m_intensityLineEdit->clear();
        m_wavelengthEditor->clearItem();
        m_inclinationEditor->clearItem();
        m_footprintEditor->clearItem();
    }
}

void SpecularBeamEditor::clearItem()
{
    m_intensityLineEdit->clear();
    m_wavelengthEditor->clearItem();
    m_inclinationEditor->clearItem();
    m_footprintEditor->clearItem();
}

//--------------------------------------------------------------------------------------------------
// private slots
//--------------------------------------------------------------------------------------------------

void SpecularBeamEditor::onIntensityEditingFinished()
{
    if (m_item)
        m_item->setIntensity(m_intensityLineEdit->text().toDouble());
}
