//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/PropertyEditor/SpecularDataPropertyWidget.cpp
//! @brief     Implements class SpecularDataPropertyWidget
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/PropertyEditor/SpecularDataPropertyWidget.h"
#include "GUI/Model/Data/AxesItems.h"
#include "GUI/Model/Data/Data1DViewItem.h"
#include "GUI/Model/Data/SpecularDataItem.h"
#include "GUI/Util/ComboProperty.h"
#include "GUI/View/Edit/DoubleSpinBox.h"
#include "GUI/View/Tool/LayoutUtils.h"
#include <QAction>
#include <QCheckBox>
#include <QComboBox>
#include <QFormLayout>
#include <QGroupBox>
#include <QLineEdit>

SpecularDataPropertyWidget::SpecularDataPropertyWidget(QWidget* parent)
    : QWidget(parent), m_item(nullptr)
{
    setWindowTitle("Properties");
    setAttribute(Qt::WA_StyledBackground, true);
    setProperty("stylable", true); // for stylesheet addressing

    m_mainLayout = new QFormLayout(this);
    m_mainLayout->setContentsMargins(8, 20, 8, 8);
    m_mainLayout->setSpacing(5);
}

void SpecularDataPropertyWidget::setItem(SpecularDataItem* item)
{
    setCurrentItem(item);
}

void SpecularDataPropertyWidget::setItem(Data1DViewItem* item)
{
    setCurrentItem(item);
}

void SpecularDataPropertyWidget::setCurrentItem(SessionItem* item)
{
    if (m_item)
        m_item->mapper()->unsubscribe(this);

    GUI::Util::Layout::clearLayout(m_mainLayout);
    m_updaters.clear();
    m_item = item;
    if (!m_item)
        return;

    BasicAxisItem* xAxisItem = nullptr;
    AmplitudeAxisItem* yAxisItem = nullptr;
    SelectionDescriptor<QString> axesUnitsDescriptor;

    if (auto* p = dynamic_cast<SpecularDataItem*>(m_item)) {
        axesUnitsDescriptor = p->axesUnits();
        xAxisItem = p->xAxisItem();
        yAxisItem = p->yAxisItem();
    } else if (auto* p = dynamic_cast<Data1DViewItem*>(m_item)) {
        axesUnitsDescriptor = p->axesUnitsDescriptor();
        xAxisItem = p->xAxisItem();
        yAxisItem = p->yAxisItem();
    } else
        return;

    m_mainLayout->addRow("Axes units:", createComboBox(axesUnitsDescriptor));

    // -- x-axis
    auto* xGroup = new QGroupBox("X axis", this);
    auto* xFormLayout = new QFormLayout(xGroup);
    xFormLayout->setContentsMargins(0, 0, 0, 0);
    xFormLayout->setSpacing(5);

    xFormLayout->addRow("Min:", createDoubleSpinbox(xAxisItem->lowerBound()));
    xFormLayout->addRow("Max:", createDoubleSpinbox(xAxisItem->upperBound()));
    xFormLayout->addRow("Title:", createTextEdit(xAxisItem->titleItem()));

    m_mainLayout->addRow(xGroup);

    // -- y-axis
    auto* yGroup = new QGroupBox("Y axis", this);
    auto* yFormLayout = new QFormLayout(yGroup);
    yFormLayout->setContentsMargins(0, 0, 0, 0);
    yFormLayout->setSpacing(5);

    yFormLayout->addRow("Min:", createDoubleSpinbox(yAxisItem->lowerBound()));
    yFormLayout->addRow("Max:", createDoubleSpinbox(yAxisItem->upperBound()));
    yFormLayout->addRow("Title:", createTextEdit(yAxisItem->titleItem()));
    yFormLayout->addRow(createCheckBox("log10", yAxisItem->logScaleItem()));

    m_mainLayout->addRow(yGroup);

    updateUIValues();

    // react on external changes (e.g. zooming in customplot shall update the axis values)
    m_item->mapper()->setOnChildPropertyChange(
        [=](SessionItem*, const QString&) { updateUIValues(); }, this);
}

QWidget* SpecularDataPropertyWidget::createDoubleSpinbox(DoubleDescriptor d)
{
    auto* spinBox = new DoubleSpinBox(this, d);
    QObject::connect(spinBox, &DoubleSpinBox::baseValueChanged,
                     [=](double newValue) { d.set(newValue); });

    m_updaters << [=]() { spinBox->updateValue(); };

    return spinBox;
}

QWidget* SpecularDataPropertyWidget::createTextEdit(SessionItem* item)
{
    auto* edit = new QLineEdit(this);
    connect(edit, &QLineEdit::textEdited, [=](const QString& t) { item->setValue(t); });

    m_updaters << [=]() { edit->setText(item->value().toString()); };

    return edit;
}

QWidget* SpecularDataPropertyWidget::createCheckBox(const QString& title, SessionItem* item)
{
    return createCheckBox(
        title, [=]() { return item->value().toBool(); }, [=](bool b) { item->setValue(b); });
}

QWidget* SpecularDataPropertyWidget::createCheckBox(const QString& title, function<bool()> getter,
                                                    function<void(bool)> setter)
{
    auto* checkBox = new QCheckBox(title, this);
    connect(checkBox, &QCheckBox::stateChanged, [=]() { setter(checkBox->isChecked()); });

    m_updaters << [=]() {
        QSignalBlocker b(checkBox);
        checkBox->setChecked(getter());
    };

    return checkBox;
}

QWidget* SpecularDataPropertyWidget::createComboBox(SelectionDescriptor<QString> d)
{
    auto* combo = new QComboBox(this);
    combo->addItems(d.options);
    combo->setMaxCount(d.options.size());

    connect(combo, qOverload<int>(&QComboBox::currentIndexChanged),
            [=](int newIndex) { d.setCurrentIndex(newIndex); });

    m_updaters << [=]() {
        QSignalBlocker b(combo);
        combo->setCurrentIndex(d.currentIndex());
    };

    return combo;
}

void SpecularDataPropertyWidget::updateUIValues()
{
    for (const auto& updater : m_updaters)
        updater();
}
