//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Tests/Functional/GUI/Check.cpp
//! @brief     Implements function checkSimulation for GUI standard test
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "BABuild.h"
#include "BATesting.h"
#include "Base/Util/FileSystemUtils.h"
#include "Core/Simulation/ISimulation.h"
#include "Device/Data/DataUtils.h"
#include "Device/Histo/IntensityDataIOFactory.h"
#include "Device/Histo/SimulationResult.h"
#include "GUI/Model/From/GUIObjectBuilder.h"
#include "GUI/Model/Instrument/InstrumentItems.h"
#include "GUI/Model/Instrument/InstrumentModel.h"
#include "GUI/Model/Material/MaterialModel.h"
#include "GUI/Model/Sample/SampleModel.h"
#include "GUI/Model/Session/SimulationOptionsItem.h"
#include "GUI/Model/To/DomainSimulationBuilder.h"
#include <iostream>

namespace {

std::unique_ptr<OutputData<double>> domainData(const std::string& /*test_name*/,
                                               const ISimulation& direct_simulation)
{
    // initializing necessary GUI
    SimulationOptionsItem optionsItem;
    SampleModel sampleModel;
    InstrumentModel instrumentModel;
    MaterialModel materialModel;

    // populating GUI models from domain
    GUI::Transform::FromDomain::populateSampleModel(&sampleModel, &materialModel,
                                                    *direct_simulation.sample());
    GUI::Transform::FromDomain::populateInstrumentModel(&instrumentModel, direct_simulation);
    GUI::Transform::FromDomain::populateSimulationOptions(&optionsItem, direct_simulation);

    std::unique_ptr<ISimulation> domain_simulation =
        GUI::Model::DomainSimulationBuilder::createSimulation(
            sampleModel.multiLayerItem(), instrumentModel.instrumentItems().front(), optionsItem);

    domain_simulation->runSimulation();
    return std::unique_ptr<OutputData<double>>(domain_simulation->result().data());
}

} // namespace

//! Run simulation directly (in core) and through GUI model, and compare results.

bool checkSimulation(const std::string& name, const ISimulation& direct_simulation,
                     const double limit)
{
    const std::unique_ptr<OutputData<double>> domain_data = domainData(name, direct_simulation);

    const std::unique_ptr<OutputData<double>> ref_data = direct_simulation.result().data();

    bool ok = DataUtils::Data::checkRelativeDifference(*domain_data, *ref_data, limit);

    if (ok)
        return true;

    // Save simulation, as it differs from reference.
    BaseUtils::Filesystem::createDirectories(BATesting::TestOutDir_Std());
    std::map<const std::string, const OutputData<double>*> tosave{{"gui", domain_data.get()},
                                                                  {"std", ref_data.get()}};
    for (const auto& [kind, data] : tosave) {
        std::string out_fname = BaseUtils::Filesystem::jointPath(BATesting::TestOutDir_Std(),
                                                                 name + "." + kind + ".int.gz");
        IntensityDataIOFactory::writeOutputData(*data, out_fname);
        std::cout << "Saved " << out_fname << "\n";
    }
    std::cout << "Notes:\n"
              << "- to visualize an intensity map, use " << BABuild::srcDir()
              << "/devtools/view/plot_int.py\n"
              << "- to plot a difference image, use " << BABuild::srcDir()
              << "/devtools/view/plot_diff_int.py\n";
    return false;
}
