#include "Base/Const/Units.h"
#include "GUI/Model/From/FromDomain.h"
#include "GUI/Model/Sample/FTDistributionItems.h"
#include "GUI/Model/Sample/InterferenceItems.h"
#include "GUI/Model/Sample/Lattice2DItems.h"
#include "GUI/Model/Sample/LayerItem.h"
#include "GUI/Model/Sample/MultiLayerItem.h"
#include "GUI/Model/Sample/ParticleLayoutItem.h"
#include "GUI/Model/Sample/SampleModel.h"
#include "Sample/Aggregate/Interference2DParaCrystal.h"
#include "Tests/GTestWrapper/google_test.h"

class TestParaCrystalItems : public ::testing::Test {
};

TEST_F(TestParaCrystalItems, Para2DFromToDomain)
{
    double length1(10.0), length2(20.0), angle(45.0), xi(90.0);
    double damping_length(1000.0), domain_size1(50.0), domain_size2(100.0);

    Interference2DParaCrystal orig(
        BasicLattice2D(length1, length2, angle * Units::deg, xi * Units::deg), damping_length,
        domain_size1, domain_size2);

    double clength_x(1.0), clength_y(2.0), gamma(3.0);
    orig.setProbabilityDistributions(
        FTDistribution2DCauchy(clength_x, clength_y, gamma * Units::deg),
        FTDistribution2DGauss(clength_x, clength_y, gamma * Units::deg));

    // from domain
    Interference2DParaCrystalItem item;
    GUI::Transform::FromDomain::set2DParaCrystalItem(&item, orig);

    EXPECT_EQ(item.dampingLength(), orig.dampingLength());
    EXPECT_EQ(item.domainSize1(), orig.domainSizes()[0]);
    EXPECT_EQ(item.domainSize2(), orig.domainSizes()[1]);
    EXPECT_FALSE(orig.integrationOverXi());
    EXPECT_EQ(item.xiIntegration(), orig.integrationOverXi());

    Lattice2DItem* latticeItem = item.latticeType().currentItem();
    EXPECT_EQ(latticeItem->modelType(), BasicLattice2DItem::M_TYPE);
    auto* basLatticeItem = dynamic_cast<BasicLattice2DItem*>(latticeItem);
    EXPECT_EQ(basLatticeItem->latticeLength1(), length1);
    EXPECT_EQ(basLatticeItem->latticeLength2(), length2);
    EXPECT_EQ(basLatticeItem->latticeAngle(), angle);
    EXPECT_EQ(basLatticeItem->latticeRotationAngle(), xi);

    FTDistribution2DItem* pdfItem1 = item.probabilityDistribution1().currentItem();
    EXPECT_EQ(pdfItem1->modelType(), FTDistribution2DCauchyItem::M_TYPE);
    EXPECT_EQ(pdfItem1->omegaX(), clength_x);
    EXPECT_EQ(pdfItem1->omegaY(), clength_y);
    EXPECT_EQ(pdfItem1->gamma(), gamma);

    FTDistribution2DItem* pdfItem2 = item.probabilityDistribution2().currentItem();
    EXPECT_EQ(pdfItem2->modelType(), FTDistribution2DGaussItem::M_TYPE);
    EXPECT_EQ(pdfItem2->omegaX(), clength_x);
    EXPECT_EQ(pdfItem2->omegaY(), clength_y);
    EXPECT_EQ(pdfItem2->gamma(), gamma);

    // to domain
    auto ifun = item.createInterference();
    std::unique_ptr<Interference2DParaCrystal> domain(
        dynamic_cast<Interference2DParaCrystal*>(ifun->clone()));
    EXPECT_EQ(domain->integrationOverXi(), orig.integrationOverXi());
    EXPECT_EQ(domain->domainSizes(), orig.domainSizes());
    EXPECT_EQ(domain->dampingLength(), orig.dampingLength());
    EXPECT_EQ(domain->lattice().length1(), orig.lattice().length1());
    EXPECT_EQ(domain->lattice().length2(), orig.lattice().length2());
    EXPECT_EQ(domain->lattice().latticeAngle(), orig.lattice().latticeAngle());
    EXPECT_EQ(domain->lattice().rotationAngle(), orig.lattice().rotationAngle());
}
