#include "GUI/Model/Material/MaterialModel.h"
#include "GUI/Model/Sample/ParticleCompositionItem.h"
#include "GUI/Model/Sample/ParticleCoreShellItem.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Model/Sample/SampleModel.h"
#include "GUI/Model/Session/SessionItemUtils.h"
#include "GUI/Model/Types/DoubleDescriptor.h"
#include "GUI/Model/Types/VectorItem.h"
#include "Tests/GTestWrapper/google_test.h"
#include <QtTest>

using namespace GUI::Session::ItemUtils;

class TestParticleCoreShell : public ::testing::Test {
};

//! Checking that adding and removing core/shell leads to enabling/disabling of their position
//! and abundance properties.

TEST_F(TestParticleCoreShell, propertyAppearance)
{
    SampleModel model;
    MaterialModel materials;
    materials.addRefractiveMaterial("Default", 1e-3, 1e-5);

    // empty coreshell particle
    auto* coreshell = model.insertItem<ParticleCoreShellItem>();
    EXPECT_TRUE(coreshell->abundanceItem()->isEnabled());
    EXPECT_EQ(coreshell->abundance(), 1.0);
    EXPECT_TRUE(coreshell->positionItem()->isEnabled());
    R3 pos = coreshell->positionItem()->getVector();
    EXPECT_EQ(pos.x(), 0.0);
    EXPECT_EQ(pos.y(), 0.0);
    EXPECT_EQ(pos.z(), 0.0);

    // adding core, and checking that abundance is disabled
    auto* core = coreshell->createCore(&materials);
    EXPECT_FALSE(core->abundanceItem()->isEnabled());
    EXPECT_TRUE(core->positionItem()->isEnabled());

    // removing core, checking that abundance restored
    coreshell->takeRow(ParentRow(*core));
    EXPECT_TRUE(core->abundanceItem()->isEnabled());
    EXPECT_TRUE(core->positionItem()->isEnabled());
    delete core;

    // create a shell and checking abundance disabled
    coreshell->createShell(&materials);
    EXPECT_FALSE(coreshell->shell()->abundanceItem()->isEnabled());
    EXPECT_FALSE(coreshell->shell()->positionItem()->isEnabled());
    // checking that position and abundance values were reset to defaults
    EXPECT_EQ(coreshell->shell()->positionItem()->x(), 0.0);
    EXPECT_EQ(coreshell->shell()->abundance(), 1.0);
}


//! Checking that abundance gets disabled in particle composition context.

TEST_F(TestParticleCoreShell, compositionContext)
{
    SampleModel model;

    // coreshell particle
    auto* coreshell = model.insertItem<ParticleCoreShellItem>();
    coreshell->setAbundance(0.2);
    EXPECT_TRUE(coreshell->abundanceItem()->isEnabled());
    EXPECT_EQ(coreshell->abundance(), 0.2);

    // create composition, adding coreshell to it
    auto* composition = model.insertItem<ParticleCompositionItem>();
    composition->addParticle(coreshell);
    // checking abundance has switched to defaults
    EXPECT_FALSE(coreshell->abundanceItem()->isEnabled());
    EXPECT_EQ(coreshell->abundance(), 1.0);

    // removing coreshell
    composition->takeRow(ParentRow(*coreshell));
    EXPECT_TRUE(coreshell->abundanceItem()->isEnabled());
    delete coreshell;
}
